<!---

https://edward.id

--->
<!doctype html>
<meta name="viewport" content="width=device-width">
<meta http-equiv="x-ua-compatible" content="ie=edge">
<title>TOTP</title>
<script>

const list = [
"totp.io test OVEK7TIJ3A3DM3M6",
];

// Reference: https://github.com/khovansky-al/web-otp-demo
if (window.crypto) {
    function generateKey(secret, counter, callback) {
        try {
            const secretBytes = base32Decode(secret);
            const counterArray = padCounter(counter);

            const hash = { name: 'HMAC', hash: { name: 'SHA-1' } };
            window.crypto.subtle.importKey('raw', secretBytes, hash, false, ['sign']).then(function(key) {
                window.crypto.subtle.sign(hash.name, key, counterArray).then(function(key) {
                    callback(new Uint8Array(key));
                });
            });
        } catch (err) { callback(new Uint8Array()); }
    }
}
else if (window.msCrypto) {
    function generateKey(secret, counter, callback) {
        try {
            const secretBytes = base32Decode(secret);
            const counterArray = padCounter(counter);

            const hash = { name: 'HMAC', hash: { name: 'SHA-1' } };
            window.msCrypto.subtle.importKey('raw', secretBytes, hash, false, ['sign']).oncomplete = function(e) {
                window.msCrypto.subtle.sign(hash, e.target.result, counterArray).oncomplete = function(e) {
                    callback(new Uint8Array(e.target.result));
                };
            };
        } catch (err) { callback(new Uint8Array()); }
    }
}
else {
    var error = true;
    alert('Error: no crypto!');
}

function base32Decode(str) {
    const base32chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
    let bits = '';
    for (let i = 0; i < str.length; i++) {
        const value = base32chars.indexOf(str.charAt(i).toUpperCase());
        if (value == -1) continue;
        const binary = value.toString(2);
        const padding = '00000'.substr(binary.length);
        bits += padding + binary;
    }
    let output = new Uint8Array(bits.length / 8);
    for (let i = 0; i < bits.length / 8; i++) {
        output[i] = parseInt(bits.substr(i * 8, 8), 2);
    }
    return output;
}

function padCounter(counter) {
    const buffer = new ArrayBuffer(8);
    const bView = new DataView(buffer);
    const byteString = '0000000000000000000000000000000000000000000000000000000000000000';
    const bCounter = (byteString + counter.toString(2)).slice(-64);
    for (let byte = 0; byte < 64; byte += 8) {
        const byteValue = parseInt(bCounter.slice(byte, byte + 8), 2);
        bView.setUint8(byte / 8, byteValue);
    }
    return buffer;
}

function DT(HS) {
    const offset = HS[19] & 15;
    const P = ((HS[offset] & 127) << 24) | (HS[offset + 1] << 16) | (HS[offset + 2] << 8) | HS[offset + 3];
    return P.toString(2);
}

function generateHOTP(secret, counter, callback) {
    generateKey(secret, counter, function(key) {
        const truncated = parseInt(DT(key), 2);
        const padded = ('000000' + (truncated % 1000000)).slice(-6);
        callback(padded);
    });
}

function generateTOTP(secret, callback) {
    generateHOTP(secret, getCounter(), callback);
}

function getCounter() {
    return Math.floor(Date.now() / 30000);
}

function getTimeLeft() {
    const timeSinceStep = Date.now() - getCounter() * 30000;
    const timeLeft = Math.ceil((30000 - timeSinceStep) / 1000);
    return timeLeft;
}

function drawArc(ctx, centerX, centerY, radius, startAngle, endAngle, globalCompositeOperation, fillStyle) {
    ctx.beginPath();
    ctx.moveTo(centerX, centerY);
    ctx.arc(centerX, centerY, radius, startAngle, endAngle);
    ctx.globalCompositeOperation = globalCompositeOperation;
    ctx.fillStyle = fillStyle;
    ctx.fill();
}

function refreshGraphic(id, timeLeft) {
    const ctx = document.getElementById('timeleft2_' + id).getContext('2d');
    ctx.clearRect(0, 0, 80, 80);
    drawArc(ctx, 40, 40, 40, 0, 2*Math.PI, 'source-over', '#E6E6E6');
    if (timeLeft == 30) {
        drawArc(ctx, 40, 40, 40, 0, 2*Math.PI, 'source-over', '#4F81BD');
    }
    else {
        drawArc(ctx, 40, 40, 40, -Math.PI/2 + ((30 - timeLeft) / 30 * 2*Math.PI), -Math.PI/2, 'source-over', '#4F81BD');
    }
    drawArc(ctx, 40, 40, 32, 0, 2*Math.PI, 'destination-out');
}

function refreshEntry(id, secret) {
    generateTOTP(secret, function(totp) {
        const timeLeft = getTimeLeft();
        document.getElementById('totp_' + id).innerHTML = '<span>' + totp.substr(0, 3) + '</span>';
        document.getElementById('totp_' + id).innerHTML += '<span style="padding-left:0.2em;">' + totp.substr(3) + '</span>';
        document.getElementById('timeleft_' + id).innerHTML = timeLeft;
        refreshGraphic(id, timeLeft);
    });
}

function refreshTable() {
    for (let i = 0; i < list.length; i++) {
        refreshEntry(i, list[i].split(' ').pop());
    }
}

function writeTable() {
    let output = '<table cellspacing=0 style="width:100%; min-width:200pt; max-width:500pt; margin:0 auto;">';
    for (let i = 0; i < list.length; i++) {
        const tokens = list[i].split(' ');
        tokens.pop();
        output += '<tr><td style="width:100%;">' + tokens.join(' ') + '</td>';
        output += '<td id="' + 'totp_' + i + '" style="padding-left:4.5em; font-size:14pt;"></td>';
        output += '<td style="padding-left:1.5em;">';
        output += '<span id="' + 'timeleft_' + i + '" style="position:absolute; width:15pt; height:15pt; line-height:15pt; font-size:7pt; font-weight:bold; text-align:center;"></span>';
        output += '<canvas id="' + 'timeleft2_' + i + '" width=80 height=80 style="width:15pt; height:15pt;"></canvas></td></tr>';
    }
    output += '<tr><td colspan=3 align=center onclick="addEntry();">+</td></tr>';
    output += '</table>';
    document.getElementById('output').innerHTML = output;
}

function addEntry() {
    const entry = prompt('Add temporary entry:', '');
    if(!entry) return;
    list.push(entry);
    writeTable();
    refreshTable();
}

window.onkeyup = function(e) {
    if (e.key == 'Add' || e.key == '+') addEntry();
}

window.onload = function() {
    if (error) return;
    writeTable();
    refreshTable();
    setInterval("refreshTable()", 500);
}

const fontSize = window.getComputedStyle(document.documentElement).fontSize;
const scale = (fontSize.substr(2) == 'px') ? parseInt(fontSize) / 16 : 1;

window.resizeTo(370 * scale, 500 * scale);

</script>

<style>
body { font-family:sans-serif; }
tr:hover { background-color: #F7F7F7; }
td { padding:0.7em; border-bottom:1px dashed #CCC; }
</style>

<div id="output" style="padding:0.5em;"></div>
